/* fso-gpsd - a gpsd replacement which connects to the fso frameworkd
 *
 * Copyright (C) 2008 Sascha Wessel <wessel@nefkom.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <math.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <glib-object.h>
#include <glib.h>

#define KNOTS_TO_KPH (1.852)
#define KNOTS_TO_MPS (0.51444444)

/******************************************************************************/

#define NAME	"fso-gpsd"
#define PORT	"2947"
#define OUSAGED "org.freesmartphone.ousaged"
#define OGPSD	"org.freesmartphone.ogpsd"
#define USAGE	"/org/freesmartphone/Usage"
#define GYPSY	"/org/freedesktop/Gypsy"

/******************************************************************************/

struct client {
	GIOChannel *channel;
	gboolean watcher;
	gboolean raw;
};

struct satellite {
	unsigned prn;
	gboolean used;
	unsigned elevation;
	unsigned azimuth;
	unsigned snr;
};

/******************************************************************************/

static gboolean nowait = FALSE;

static gboolean info_queued = FALSE;
static gboolean sats_queued = FALSE;
static gboolean nmea_queued = FALSE;

/******************************************************************************/

static GList *client_list = NULL;

/******************************************************************************/

static DBusGConnection *connection;
static DBusGProxy *proxy_usage;
static DBusGProxy *proxy_device;
static DBusGProxy *proxy_position;
static DBusGProxy *proxy_accuracy;
static DBusGProxy *proxy_course;
static DBusGProxy *proxy_satellite;
static DBusGProxy *proxy_time;

/******************************************************************************/

static gboolean device_connectionstatus;
static int device_fixstatus; /* 0=INVALID, 1=NONE, 2=2D, 3=3D */

static int position_timestamp;
static double position_latitude;
static double position_longitude;
static double position_altitude;

static double accuracy_pdop;
static double accuracy_hdop;
static double accuracy_vdop;

static int course_timestamp;
static double course_speed;
static double course_direction;
static double course_climb;

#define SATELLITEC_MAX 32
static struct satellite satellite_satellitev[SATELLITEC_MAX];
static unsigned satellite_satellitec;

static int time_seconds;

/******************************************************************************/

static GString *nmea_gga;
static GString *nmea_gsa;
static GString *nmea_gsv[3];
static GString *nmea_rmc;
static GString *nmea_gll;
static GString *nmea_vtg;

/******************************************************************************/

static int
get_fixstatus(void)
{
	if (!isnan(position_latitude) && !isnan(position_longitude)) {
		if (!isnan(position_altitude)) {
			return MIN(device_fixstatus, 3);
		} else {
			return MIN(device_fixstatus, 2);
		}
	} else {
		return MIN(device_fixstatus, 1);
	}
}

/******************************************************************************/

static void
shutdown_client(struct client *client)
{
	if (client->channel) {
		g_message("Shutdown client");

		g_io_channel_shutdown(client->channel, FALSE, NULL);
	}

	client->channel = NULL;
	client->watcher = FALSE;
	client->raw = FALSE;
}

static void
send_to_client(struct client *client, const char *string, gsize len)
{
	GIOStatus status;
	gsize bytes_written;

	if (!client->channel) {
		return;
	}

	status = g_io_channel_write_chars(client->channel, string,
		len, &bytes_written, NULL);

	if ((bytes_written != len) || (status != G_IO_STATUS_NORMAL)) {
		shutdown_client(client);
	}
}

/******************************************************************************/

static void
add_nmea_checksum(GString *nmea)
{
	int checksum = 0;
	char *p;

	p = nmea->str;

	if (*p == '$') {
		p++;
	}

	while (*p) {
		checksum ^= (int) *p;
		p++;
	}

	g_string_append_printf(nmea, "*%02x\r\n", checksum);
}

static void
add_nmea_time(GString *nmea)
{
	if (time_seconds <= 0) {
		g_string_append_c(nmea, ',');
	} else {
		struct tm *tm;
		time_t seconds;

		seconds = time_seconds;
		tm = gmtime(&seconds);

		g_string_append_printf(nmea, ",%02d%02d%02d.00",
			tm->tm_hour, tm->tm_min, tm->tm_sec);
	}
}

static void
add_nmea_latitude(GString *nmea)
{
	if (isnan(position_latitude)) {
		g_string_append(nmea, ",,");
	} else {
		double a, d;

		a = fabs(position_latitude);
		d = floor(a);

		g_string_append_printf(nmea, ",%010.5f,%C",
			(d * 100.0) + ((a - d) * 60.0),
			(position_latitude > 0) ? 'N' : 'S');
	}
}

static void
add_nmea_longitude(GString *nmea)
{
	if (isnan(position_longitude)) {
		g_string_append(nmea, ",,");
	} else {
		double a, d;

		a = fabs(position_longitude);
		d = floor(a);

		g_string_append_printf(nmea, ",%011.5f,%C",
			(d * 100.0) + ((a - d) * 60.0),
			(position_longitude > 0) ? 'E' : 'W');
	}
}

/******************************************************************************/

static void
update_nmea_gga(void)
{
	unsigned i, used;

	g_string_assign(nmea_gga, "$GPGGA");

	add_nmea_time(nmea_gga);

	add_nmea_latitude(nmea_gga);

	add_nmea_longitude(nmea_gga);

	g_string_append_printf(nmea_gga, ",%1d", MAX(get_fixstatus() - 1, 0));

	for (i = 0, used = 0; i < satellite_satellitec; i++) {
		if (satellite_satellitev[i].used) {
			used++;
		}
	}

	g_string_append_printf(nmea_gga, ",%02d", used);

	if (isnan(accuracy_hdop)) {
		g_string_append_c(nmea_gga, ',');
	} else {
		g_string_append_printf(nmea_gga, ",%04.2f", accuracy_hdop);
	}

	if (isnan(position_altitude)) {
		g_string_append(nmea_gga, ",,");
	} else {
		g_string_append_printf(nmea_gga, ",%03.1f,M",
			position_altitude);
	}

	g_string_append(nmea_gga, ",,"); /* FIXME */

	g_string_append(nmea_gga, ",,");

	add_nmea_checksum(nmea_gga);
}

static void
update_nmea_gsa(void)
{
	unsigned i;

	g_string_assign(nmea_gsa, "$GPGSA");

	g_string_append(nmea_gsa, ",A");

	g_string_append_printf(nmea_gsa, ",%1d", MAX(get_fixstatus() - 1, 0));

	for (i = 0; i < 12; i++) {
		if (i < satellite_satellitec) {
			g_string_append_printf(nmea_gsa, ",%02d",
				satellite_satellitev[i].prn);
		} else {
			g_string_append_c(nmea_gsa, ',');
		}
	}

	if (isnan(accuracy_pdop)) {
		g_string_append_c(nmea_gsa, ',');
	} else {
		g_string_append_printf(nmea_gsa, ",%04.2f", accuracy_pdop);
	}

	if (isnan(accuracy_hdop)) {
		g_string_append_c(nmea_gsa, ',');
	} else {
		g_string_append_printf(nmea_gsa, ",%04.2f", accuracy_hdop);
	}

	if (isnan(accuracy_vdop)) {
		g_string_append_c(nmea_gsa, ',');
	} else {
		g_string_append_printf(nmea_gsa, ",%04.2f", accuracy_vdop);
	}

	add_nmea_checksum(nmea_gsa);
}

static void
update_nmea_gsv(void)
{
	unsigned i;
	unsigned sat_count;
	unsigned sentence_count;

	g_string_truncate(nmea_gsv[0], 0);
	g_string_truncate(nmea_gsv[1], 0);
	g_string_truncate(nmea_gsv[2], 0);

	if (satellite_satellitec == 0) {
		g_string_assign(nmea_gsv[0], "$GPGSV,1,1,00*79\r\n");
		return;
	}

	sat_count = MIN(satellite_satellitec, 12);
	sentence_count = ((sat_count - 1) / 4) + 1;

	for (i = 0; i < sat_count; i++) {
		int index_div, index_mod;

		index_div = i / 4;
		index_mod = i % 4;

		if (index_mod == 0) {
			g_string_append_printf(nmea_gsv[index_div],
				"$GPGSV,%1d,%1d,%02d",
				sentence_count,
				index_div + 1,
				sat_count);
		}

		g_string_append_printf(nmea_gsv[index_div],
			",%02u,%02u,%03u,%02u",
			satellite_satellitev[i].prn,
			satellite_satellitev[i].elevation,
			satellite_satellitev[i].azimuth,
			satellite_satellitev[i].snr);
	}

	for (i = 0; i < 3; i++) {
		if (nmea_gsv[i]->len > 0) {
			add_nmea_checksum(nmea_gsv[i]);
		}
	}
}

static void
update_nmea_rmc(void)
{
	g_string_assign(nmea_rmc, "$GPRMC");

	add_nmea_time(nmea_rmc);

	if (get_fixstatus() > 1) {
		g_string_append(nmea_rmc, ",A");
	} else {
		g_string_append(nmea_rmc, ",V");
	}

	add_nmea_latitude(nmea_rmc);

	add_nmea_longitude(nmea_rmc);

	if (isnan(course_speed)) {
		g_string_append_c(nmea_rmc, ',');
	} else {
		g_string_append_printf(nmea_rmc, ",%07.3f",
			course_speed);
	}

	if (isnan(course_direction)) {
		g_string_append_c(nmea_rmc, ',');
	} else {
		g_string_append_printf(nmea_rmc, ",%06.2f",
			course_direction);
	}

	if (time_seconds <= 0) {
		g_string_append_c(nmea_rmc, ',');
	} else {
		struct tm *tm;
		time_t seconds;

		seconds = time_seconds;
		tm = gmtime(&seconds);

		g_string_append_printf(nmea_rmc, ",%02d%02d%02d",
			tm->tm_mday, tm->tm_mon + 1, tm->tm_year % 100);
	}

	g_string_append(nmea_rmc, ",,");

	add_nmea_checksum(nmea_rmc);
}

static void
update_nmea_gll(void)
{
	g_string_assign(nmea_gll, "$GPGLL");

	add_nmea_latitude(nmea_gll);

	add_nmea_longitude(nmea_gll);

	add_nmea_time(nmea_gll);

	if (get_fixstatus() > 1) {
		g_string_append(nmea_gll, ",A");
	} else {
		g_string_append(nmea_gll, ",V");
	}

	add_nmea_checksum(nmea_gll);
}

static void
update_nmea_vtg(void)
{
	g_string_assign(nmea_vtg, "$GPVTG");

	if (isnan(course_direction)) {
		g_string_append(nmea_vtg, ",,");
	} else {
		g_string_append_printf(nmea_vtg, ",%04.2f,T", course_direction);
	}

	g_string_append(nmea_vtg, ",,"); /* FIXME */

	if (isnan(course_speed)) {
		g_string_append(nmea_vtg, ",,,,");
	} else {
		g_string_append_printf(nmea_vtg, ",%05.3f,N,%05.3f,K",
			course_speed, course_speed * KNOTS_TO_KPH);
	}

	add_nmea_checksum(nmea_vtg);
}

/******************************************************************************/

static gboolean
send_nmea_to_clients(gpointer data)
{
	GString *string;
	GList *l;

	nmea_queued = FALSE;

	string = g_string_sized_new(1024);

	if (nmea_gga->len == 0) {
		update_nmea_gga();
	}
	if (nmea_gsa->len == 0) {
		update_nmea_gsa();
	}
	if (nmea_gsv[0]->len == 0) {
		update_nmea_gsv();
	}
	if (nmea_rmc->len == 0) {
		update_nmea_rmc();
	}
	if (nmea_gll->len == 0) {
		update_nmea_gll();
	}
	if (nmea_vtg->len == 0) {
		update_nmea_vtg();
	}

	g_string_append(string, nmea_gga->str);
	g_string_append(string, nmea_gsa->str);
	g_string_append(string, nmea_gsv[0]->str);
	g_string_append(string, nmea_gsv[1]->str);
	g_string_append(string, nmea_gsv[2]->str);
	g_string_append(string, nmea_rmc->str);
	g_string_append(string, nmea_gll->str);
	g_string_append(string, nmea_vtg->str);

	for (l = client_list; l; l = l->next) {
		struct client *client = l->data;

		if (client->raw) {
			send_to_client(client, string->str, string->len);
		}
	}

	g_string_free(string, TRUE);

	return FALSE;
}

static void
queue_send_nmea_to_clients(void)
{
	if (nmea_queued) {
		return;
	}

	nmea_queued = TRUE;

	g_idle_add(send_nmea_to_clients, NULL);
}

/******************************************************************************/

static void
add_sats(GString *string)
{
	unsigned i, count;

	if (satellite_satellitec == 0) {
		g_string_append(string, ",Y=?");
		return;
	}

	g_string_append(string, ",Y=-");

	if (time_seconds <= 0) {
		g_string_append(string, " ? ");
	} else {
		g_string_append_printf(string, " %d.000 ", time_seconds);
	}

	count = MIN(satellite_satellitec, 12);

	g_string_append_printf(string, "%d:", count);

	for (i = 0; i < count; i++) {
		g_string_append_printf(string, "%d %d %d %d %d:",
			satellite_satellitev[i].prn,
			satellite_satellitev[i].elevation,
			satellite_satellitev[i].azimuth,
			satellite_satellitev[i].snr,
			satellite_satellitev[i].used);
	}
}

/******************************************************************************/

static gboolean
send_sats_to_clients(gpointer data)
{
	GString *string;
	GList *l;

	sats_queued = FALSE;

	string = g_string_sized_new(1024);

	g_string_append(string, "GPSD");

	add_sats(string);

	g_string_append(string, "\r\n");

	for (l = client_list; l; l = l->next) {
		struct client *client = l->data;

		if (client->watcher) {
			send_to_client(client, string->str, string->len);
		}
	}

	g_string_free(string, TRUE);

	return FALSE;
}

static void
queue_send_sats_to_clients(void)
{
	if (sats_queued) {
		return;
	}

	sats_queued = TRUE;

	g_idle_add(send_sats_to_clients, NULL);
}

/******************************************************************************/

static void
add_info(GString *string)
{
	if (!device_connectionstatus || get_fixstatus() <= 1) {
		g_string_append(string, ",O=?");
		return;
	}

	g_string_append(string, ",O=-");

	if (time_seconds <= 0) {
		g_string_append(string, " ?");
	} else {
		g_string_append_printf(string, " %d.000", time_seconds);
	}

	g_string_append(string, " ?"); /* FIXME */

	if (isnan(position_latitude)) {
		g_string_append(string, " ?");
	} else {
		g_string_append_printf(string, " %.6f", position_latitude);
	}
	if (isnan(position_longitude)) {
		g_string_append(string, " ?");
	} else {
		g_string_append_printf(string, " %.6f", position_longitude);
	}
	if (isnan(position_altitude)) {
		g_string_append(string, " ?");
	} else {
		g_string_append_printf(string, " %.2f", position_altitude);
	}
	if (isnan(accuracy_hdop)) {
		g_string_append(string, " ?");
	} else {
		g_string_append_printf(string, " %.2f", accuracy_hdop);
	}
	if (isnan(accuracy_vdop)) {
		g_string_append(string, " ?");
	} else {
		g_string_append_printf(string, " %.2f", accuracy_vdop);
	}
	if (isnan(course_direction)) {
		g_string_append(string, " ?");
	} else {
		g_string_append_printf(string, " %.4f", course_direction);
	}
	if (isnan(course_speed)) {
		g_string_append(string, " ?");
	} else {
		g_string_append_printf(string, " %.3f",
			course_speed * KNOTS_TO_MPS);
	}
	if (isnan(course_climb)) {
		g_string_append(string, " ?");
	} else {
		g_string_append_printf(string, " %.3f", course_climb);
	}

	g_string_append(string, " ? ? ?"); /* FIXME */

	g_string_append_printf(string, " %d", get_fixstatus());
}

/******************************************************************************/

static gboolean
send_info_to_clients(gpointer data)
{
	GString *string;
	GList *l;

	info_queued = FALSE;

	string = g_string_sized_new(1024);

	g_string_append(string, "GPSD");

	add_info(string);

	g_string_append(string, "\r\n");

	for (l = client_list; l; l = l->next) {
		struct client *client = l->data;

		if (client->watcher) {
			send_to_client(client, string->str, string->len);
		}
	}

	g_string_free(string, TRUE);

	return FALSE;
}

static void
queue_send_info_to_clients(void)
{
	if (info_queued) {
		return;
	}

	info_queued = TRUE;

	g_idle_add(send_info_to_clients, NULL);
}

/******************************************************************************/

static gboolean
timeout_handler(gpointer data)
{
	send_info_to_clients(NULL);
	send_sats_to_clients(NULL);
	send_nmea_to_clients(NULL);

	return !!client_list;
}

/******************************************************************************/

static void
marshal_VOID__INT_INT_DOUBLE_DOUBLE_DOUBLE(
	GClosure *closure,
	GValue *return_value G_GNUC_UNUSED,
	guint n_param_values,
	const GValue *param_values,
	gpointer invocation_hint G_GNUC_UNUSED,
	gpointer marshal_data
)
{
	typedef void (*GMarshalFunc_VOID__INT_INT_DOUBLE_DOUBLE_DOUBLE)(
		gpointer data1,
		gint arg_1,
		gint arg_2,
		gdouble arg_3,
		gdouble arg_4,
		gdouble arg_5,
		gpointer data2);

	GMarshalFunc_VOID__INT_INT_DOUBLE_DOUBLE_DOUBLE callback;
	GCClosure *cc = (GCClosure*) closure;
	gpointer data1, data2;

	g_return_if_fail(n_param_values == 6);

	if (G_CCLOSURE_SWAP_DATA(closure)) {
		data1 = closure->data;
		data2 = g_value_peek_pointer(param_values + 0);
	} else {
		data1 = g_value_peek_pointer(param_values + 0);
		data2 = closure->data;
	}

	callback = (GMarshalFunc_VOID__INT_INT_DOUBLE_DOUBLE_DOUBLE) (
		marshal_data ? marshal_data : cc->callback);

	callback(data1,
		g_value_get_int(param_values + 1),
		g_value_get_int(param_values + 2),
		g_value_get_double(param_values + 3),
		g_value_get_double(param_values + 4),
		g_value_get_double(param_values + 5),
		data2);
}

static void
marshal_VOID__INT_DOUBLE_DOUBLE_DOUBLE(
	GClosure *closure,
	GValue *return_value G_GNUC_UNUSED,
	guint n_param_values,
	const GValue *param_values,
	gpointer invocation_hint G_GNUC_UNUSED,
	gpointer marshal_data
)
{
	typedef void (*GMarshalFunc_VOID__INT_DOUBLE_DOUBLE_DOUBLE)(
		gpointer data1,
		gint arg_1,
		gdouble arg_2,
		gdouble arg_3,
		gdouble arg_4,
		gpointer data2);

	GMarshalFunc_VOID__INT_DOUBLE_DOUBLE_DOUBLE callback;
	GCClosure *cc = (GCClosure*) closure;
	gpointer data1, data2;

	g_return_if_fail(n_param_values == 5);

	if (G_CCLOSURE_SWAP_DATA(closure)) {
		data1 = closure->data;
		data2 = g_value_peek_pointer(param_values + 0);
	} else {
		data1 = g_value_peek_pointer(param_values + 0);
		data2 = closure->data;
	}

	callback = (GMarshalFunc_VOID__INT_DOUBLE_DOUBLE_DOUBLE) (
		marshal_data ? marshal_data : cc->callback);

	callback(data1,
		g_value_get_int(param_values + 1),
		g_value_get_double(param_values + 2),
		g_value_get_double(param_values + 3),
		g_value_get_double(param_values + 4),
		data2);
}

/******************************************************************************/

static void
connection_status_changed(
	DBusGProxy *proxy,
	gboolean connectionstatus,
	gpointer user_data
)
{
	device_connectionstatus = connectionstatus;

	g_message("ConnectionStatusChanged(connectionstatus=%s)",
		device_connectionstatus ? "TRUE" : "FALSE");

	if ((client_list || nowait) && !device_connectionstatus) {
		dbus_g_proxy_call_no_reply(proxy_usage, "RequestResource",
			G_TYPE_STRING, "GPS",
			G_TYPE_INVALID,
			G_TYPE_INVALID);
	}
}

static void
fix_status_changed(
	DBusGProxy *proxy,
	int fixstatus,
	gpointer user_data
)
{
	const char *str;

	device_fixstatus = fixstatus;

	switch (device_fixstatus) {
	case 1: str = "NONE"; break;
	case 2: str = "2D"; break;
	case 3: str = "3D"; break;
	default: str = "INVALID"; break;
	}

	g_message("FixStatusChanged(fixstatus=%s)", str);

	g_string_truncate(nmea_gga, 0);
	g_string_truncate(nmea_gsa, 0);
	g_string_truncate(nmea_rmc, 0);
	g_string_truncate(nmea_gll, 0);

	queue_send_nmea_to_clients();
	queue_send_info_to_clients();
}

static void
position_changed(
	DBusGProxy *proxy,
	int fields,
	int timestamp,
	double latitude,
	double longitude,
	double altitude,
	gpointer user_data
)
{
	position_timestamp = timestamp;
	position_latitude  = (fields & (1 << 0)) ? latitude  : NAN;
	position_longitude = (fields & (1 << 1)) ? longitude : NAN;
	position_altitude  = (fields & (1 << 2)) ? altitude  : NAN;

	g_debug("PositionChanged(timestamp=%d, latitude=%f, longitude=%f, altitude=%f)",
		position_timestamp, position_latitude,
		position_longitude, position_altitude);

	g_string_truncate(nmea_gga, 0);
	g_string_truncate(nmea_rmc, 0);
	g_string_truncate(nmea_gll, 0);

	queue_send_nmea_to_clients();
	queue_send_info_to_clients();
}

static void
accuracy_changed(
	DBusGProxy *proxy,
	int fields,
	double pdop,
	double hdop,
	double vdop,
	gpointer user_data
)
{
	accuracy_pdop = (fields & (1 << 0)) ? pdop : NAN;
	accuracy_hdop = (fields & (1 << 1)) ? hdop : NAN;
	accuracy_vdop = (fields & (1 << 2)) ? vdop : NAN;

	g_debug("AccuracyChanged(pdop=%f, hdop=%f, vdop=%f)",
		accuracy_pdop, accuracy_hdop, accuracy_vdop);

	g_string_truncate(nmea_gga, 0);
	g_string_truncate(nmea_gsa, 0);

	queue_send_nmea_to_clients();
	queue_send_info_to_clients();
}

static void
course_changed(
	DBusGProxy *proxy,
	int fields,
	int timestamp,
	double speed,
	double direction,
	double climb,
	gpointer user_data
)
{
	course_timestamp = timestamp;
	course_speed     = (fields & (1 << 0)) ? speed     : NAN;
	course_direction = (fields & (1 << 1)) ? direction : NAN;
	course_climb     = (fields & (1 << 2)) ? climb     : NAN;

	g_debug("CourseChanged(timestamp=%d, speed=%f, direction=%f, climb=%f)",
		course_timestamp, course_speed,
		course_direction, course_climb);

	g_string_truncate(nmea_rmc, 0);
	g_string_truncate(nmea_vtg, 0);

	queue_send_nmea_to_clients();
	queue_send_info_to_clients();
}

static void
satellites_changed(
	DBusGProxy *proxy,
	GPtrArray *satellites,
	gpointer user_data
)
{
	int i;

	if (!satellites) {
		satellite_satellitec = 0;
	} else {
		satellite_satellitec = MIN(SATELLITEC_MAX, satellites->len);
	}

	for (i = 0; i < satellite_satellitec; i++) {
		struct satellite *sat;
		GValueArray *val;

		sat = satellite_satellitev + i;
		val = satellites->pdata[i];

		sat->prn = g_value_get_uint(g_value_array_get_nth(val, 0));
		sat->used = g_value_get_boolean(g_value_array_get_nth(val, 1));
		sat->elevation = g_value_get_uint(g_value_array_get_nth(val, 2));
		sat->azimuth = g_value_get_uint(g_value_array_get_nth(val, 3));
		sat->snr = g_value_get_uint(g_value_array_get_nth(val, 4));

		g_debug("SatellitesChanged(index=%d, prn=%u, used=%s, elevation=%u, azimuth=%u, snr=%u)",
			i,
			sat->prn,
			sat->used ? "TRUE" : "FALSE",
			sat->elevation,
			sat->azimuth,
			sat->snr);
	}

	g_string_truncate(nmea_gga, 0);
	g_string_truncate(nmea_gsa, 0);
	g_string_truncate(nmea_gsv[0], 0);
	g_string_truncate(nmea_gsv[1], 0);
	g_string_truncate(nmea_gsv[2], 0);

	queue_send_nmea_to_clients();
	queue_send_sats_to_clients();
}

static void
time_changed(
	DBusGProxy *proxy,
	int seconds,
	gpointer user_data
)
{
	time_seconds = seconds;

	g_debug("TimeChanged(seconds=%d)", time_seconds);

	g_string_truncate(nmea_gga, 0);
	g_string_truncate(nmea_rmc, 0);
	g_string_truncate(nmea_gll, 0);

	queue_send_nmea_to_clients();
	queue_send_info_to_clients();
}

/******************************************************************************/

static void
start(void)
{
	dbus_g_proxy_connect_signal(proxy_device, "ConnectionStatusChanged",
		G_CALLBACK(connection_status_changed), NULL, NULL);
	dbus_g_proxy_connect_signal(proxy_device, "FixStatusChanged",
		G_CALLBACK(fix_status_changed), NULL, NULL);
	dbus_g_proxy_connect_signal(proxy_position, "PositionChanged",
		G_CALLBACK(position_changed), NULL, NULL);
	dbus_g_proxy_connect_signal(proxy_accuracy, "AccuracyChanged",
		G_CALLBACK(accuracy_changed), NULL, NULL);
	dbus_g_proxy_connect_signal(proxy_course, "CourseChanged",
		G_CALLBACK(course_changed), NULL, NULL);
	dbus_g_proxy_connect_signal(proxy_satellite, "SatellitesChanged",
		G_CALLBACK(satellites_changed), NULL, NULL);
	dbus_g_proxy_connect_signal(proxy_time, "TimeChanged",
		G_CALLBACK(time_changed), NULL, NULL);

	dbus_g_proxy_call_no_reply(proxy_usage, "RequestResource",
		G_TYPE_STRING, "GPS",
		G_TYPE_INVALID,
		G_TYPE_INVALID);

	{
		gboolean connectionstatus = FALSE;

		dbus_g_proxy_call(proxy_device, "GetConnectionStatus", NULL,
			G_TYPE_INVALID,
			G_TYPE_BOOLEAN, &connectionstatus,
			G_TYPE_INVALID);
		connection_status_changed(NULL, connectionstatus, NULL);
	}
	{
		int fixstatus = 0;

		dbus_g_proxy_call(proxy_device, "GetFixStatus", NULL,
			G_TYPE_INVALID,
			G_TYPE_INT, &fixstatus,
			G_TYPE_INVALID);
		fix_status_changed(NULL, fixstatus, NULL);
	}
	{
		int fields = 0, timestamp = 0;
		double latitude = NAN, longitude = NAN, altitude = NAN;

		dbus_g_proxy_call(proxy_position, "GetPosition", NULL,
			G_TYPE_INVALID,
			G_TYPE_INT, &fields,
			G_TYPE_INT, &timestamp,
			G_TYPE_DOUBLE, &latitude,
			G_TYPE_DOUBLE, &longitude,
			G_TYPE_DOUBLE, &altitude,
			G_TYPE_INVALID);
		position_changed(NULL, fields, timestamp, latitude, longitude,
			altitude, NULL);
	}
	{
		int fields = 0;
		double pdop = NAN, hdop = NAN, vdop = NAN;

		dbus_g_proxy_call(proxy_accuracy, "GetAccuracy", NULL,
			G_TYPE_INVALID,
			G_TYPE_INT, &fields,
			G_TYPE_DOUBLE, &pdop,
			G_TYPE_DOUBLE, &hdop,
			G_TYPE_DOUBLE, &vdop,
			G_TYPE_INVALID);
		accuracy_changed(NULL, fields, pdop, hdop, vdop, NULL);
	}
	{
		int fields = 0, timestamp = 0;
		double speed = NAN, direction = NAN, climb = NAN;

		dbus_g_proxy_call(proxy_course, "GetCourse", NULL,
			G_TYPE_INVALID,
			G_TYPE_INT, &fields,
			G_TYPE_INT, &timestamp,
			G_TYPE_DOUBLE, &speed,
			G_TYPE_DOUBLE, &direction,
			G_TYPE_DOUBLE, &climb,
			G_TYPE_INVALID);
		course_changed(NULL, fields, timestamp, speed, direction,
			climb, NULL);
	}
	{
		GPtrArray *satellites = NULL;

		dbus_g_proxy_call(proxy_satellite, "GetSatellites", NULL,
			G_TYPE_INVALID,
			dbus_g_type_get_collection("GPtrArray",
				dbus_g_type_get_struct("GValueArray",
					G_TYPE_UINT,
					G_TYPE_BOOLEAN,
					G_TYPE_UINT,
					G_TYPE_UINT,
					G_TYPE_UINT,
					G_TYPE_INVALID)),
			&satellites,
			G_TYPE_INVALID);
		satellites_changed(NULL, satellites, NULL);
	}
	{
		int seconds;

		dbus_g_proxy_call(proxy_time, "GetTime", NULL,
			G_TYPE_INVALID,
			G_TYPE_INT, &seconds,
			G_TYPE_INVALID);
		time_changed(NULL, seconds, NULL);
	}
}

static void
stop(void)
{
	dbus_g_proxy_call_no_reply(proxy_usage, "ReleaseResource",
		G_TYPE_STRING, "GPS",
		G_TYPE_INVALID,
		G_TYPE_INVALID);

	dbus_g_proxy_disconnect_signal(proxy_device, "ConnectionStatusChanged",
		G_CALLBACK(connection_status_changed), NULL);
	dbus_g_proxy_disconnect_signal(proxy_device, "FixStatusChanged",
		G_CALLBACK(fix_status_changed), NULL);
	dbus_g_proxy_disconnect_signal(proxy_position, "PositionChanged",
		G_CALLBACK(position_changed), NULL);
	dbus_g_proxy_disconnect_signal(proxy_accuracy, "AccuracyChanged",
		G_CALLBACK(accuracy_changed), NULL);
	dbus_g_proxy_disconnect_signal(proxy_course, "CourseChanged",
		G_CALLBACK(course_changed), NULL);
	dbus_g_proxy_disconnect_signal(proxy_satellite, "SatellitesChanged",
		G_CALLBACK(satellites_changed), NULL);
	dbus_g_proxy_disconnect_signal(proxy_time, "TimeChanged",
		G_CALLBACK(time_changed), NULL);
}

/******************************************************************************/

static void
init_dbus(void)
{
	GError *error = NULL;

	connection = dbus_g_bus_get(DBUS_BUS_SYSTEM, &error);

	if (!connection) {
		g_printerr("Failed to open connection to system bus: %s\n",
			error->message);
		g_error_free(error);
		exit(EXIT_FAILURE);
	}

	dbus_g_object_register_marshaller(
		marshal_VOID__INT_DOUBLE_DOUBLE_DOUBLE,
		G_TYPE_NONE,
		G_TYPE_INT,
		G_TYPE_DOUBLE,
		G_TYPE_DOUBLE,
		G_TYPE_DOUBLE,
		G_TYPE_INVALID);
	dbus_g_object_register_marshaller(
		marshal_VOID__INT_INT_DOUBLE_DOUBLE_DOUBLE,
		G_TYPE_NONE,
		G_TYPE_INT,
		G_TYPE_INT,
		G_TYPE_DOUBLE,
		G_TYPE_DOUBLE,
		G_TYPE_DOUBLE,
		G_TYPE_INVALID);

	proxy_usage = dbus_g_proxy_new_for_name(connection, OUSAGED, USAGE,
		"org.freesmartphone.Usage");
	proxy_device = dbus_g_proxy_new_for_name(connection, OGPSD, GYPSY,
		"org.freedesktop.Gypsy.Device");
	proxy_position = dbus_g_proxy_new_for_name(connection, OGPSD, GYPSY,
		"org.freedesktop.Gypsy.Position");
	proxy_accuracy = dbus_g_proxy_new_for_name(connection, OGPSD, GYPSY,
		"org.freedesktop.Gypsy.Accuracy");
	proxy_course = dbus_g_proxy_new_for_name(connection, OGPSD, GYPSY,
		"org.freedesktop.Gypsy.Course");
	proxy_satellite = dbus_g_proxy_new_for_name(connection, OGPSD, GYPSY,
		"org.freedesktop.Gypsy.Satellite");
	proxy_time = dbus_g_proxy_new_for_name(connection, OGPSD, GYPSY,
		"org.freedesktop.Gypsy.Time");

	dbus_g_proxy_add_signal(proxy_device, "ConnectionStatusChanged",
		G_TYPE_BOOLEAN,
		G_TYPE_INVALID);
	dbus_g_proxy_add_signal(proxy_device, "FixStatusChanged",
		G_TYPE_INT,
		G_TYPE_INVALID);
	dbus_g_proxy_add_signal(proxy_position, "PositionChanged",
		G_TYPE_INT,
		G_TYPE_INT,
		G_TYPE_DOUBLE,
		G_TYPE_DOUBLE,
		G_TYPE_DOUBLE,
		G_TYPE_INVALID);
	dbus_g_proxy_add_signal(proxy_accuracy, "AccuracyChanged",
		G_TYPE_INT,
		G_TYPE_DOUBLE,
		G_TYPE_DOUBLE,
		G_TYPE_DOUBLE,
		G_TYPE_INVALID);
	dbus_g_proxy_add_signal(proxy_course, "CourseChanged",
		G_TYPE_INT,
		G_TYPE_INT,
		G_TYPE_DOUBLE,
		G_TYPE_DOUBLE,
		G_TYPE_DOUBLE,
		G_TYPE_INVALID);
	dbus_g_proxy_add_signal(proxy_satellite, "SatellitesChanged",
		dbus_g_type_get_collection("GPtrArray",
			dbus_g_type_get_struct("GValueArray",
				G_TYPE_UINT,
				G_TYPE_BOOLEAN,
				G_TYPE_UINT,
				G_TYPE_UINT,
				G_TYPE_UINT,
				G_TYPE_INVALID)),
		G_TYPE_INVALID);
	dbus_g_proxy_add_signal(proxy_time, "TimeChanged",
		G_TYPE_INT,
		G_TYPE_INVALID);
}

/******************************************************************************/

static gboolean
read_from_client(GIOChannel *source, GIOCondition condition, gpointer data)
{
	struct client *client;
	char buf[4096];
	gsize len = sizeof(buf);
	GIOStatus status;
	GString *reply;
	char *p;

	client = data;

	if (condition != G_IO_IN) {
		shutdown_client(client);
	}

	if (!client->channel) {
		g_message("Removing client");

		client_list = g_list_remove(client_list, client);
		g_free(client);
		g_io_channel_unref(source);

		if (!client_list && !nowait) {
			stop();
		}

		return FALSE;
	}

	status = g_io_channel_read_chars(source, buf, len - 2, &len, NULL);

	if ((status != G_IO_STATUS_NORMAL) || (len < 1)) {
		shutdown_client(client);
		return TRUE;
	}

	if (buf[len - 1] != '\n') {
		buf[len] = '\n';
		len++;
	}

	buf[len] = '\0';

	reply = g_string_sized_new(128);
	g_string_assign(reply, "GPSD");

	p = buf;

	while (*p) {
		switch (toupper(*p++)) {
		case 'A':
			if (isnan(position_altitude)) {
				g_string_append(reply, ",A=?");
			} else {
				g_string_append_printf(reply, ",A=%.3f",
					position_altitude);
			}
			break;
		case 'B':
			g_string_append(reply, ",B=?");
			break;
		case 'C':
			g_string_append(reply, ",C=1.00"); /* FIXME */
			break;
		case 'D':
			if (time_seconds <= 0) {
				g_string_append(reply, ",D=?");
			} else {
				struct tm *tm;
				time_t seconds;

				seconds = time_seconds;
				tm = gmtime(&seconds);

				g_string_append_printf(reply,
					",D=%04d-%02d-%02dT%02d:%02d:%02d.00Z",
					tm->tm_year + 1900, tm->tm_mon + 1,
					tm->tm_mday, tm->tm_hour, tm->tm_min,
					tm->tm_sec);
			}
			break;
		case 'E':
			g_string_append(reply, ",E=");

			if (isnan(accuracy_pdop)) {
				g_string_append(reply, "?");
			} else {
				g_string_append_printf(reply, "%.2f",
					accuracy_pdop);
			}
			if (isnan(accuracy_hdop)) {
				g_string_append(reply, " ?");
			} else {
				g_string_append_printf(reply, " %.2f",
					accuracy_hdop);
			}
			if (isnan(accuracy_vdop)) {
				g_string_append(reply, " ?");
			} else {
				g_string_append_printf(reply, " %.2f",
					accuracy_vdop);
			}
			break;
		case 'F':
			if (*p == '=') {
				while (isprint(*p) && !isspace(*p)) {
					p++;
				}
			}
			g_string_append(reply, ",F=gypsy");
			break;
		case 'G':
			if (*p == '=') {
				p += strcspn(p, ",\r\n");
			}
			g_string_append(reply, ",G=GPS");
			break;
		case 'I':
			g_string_append(reply, ",I=gypsy");
			break;
		case 'J':
			if (*p == '=') {
				p++;
			}
			if (*p == '1' || *p == '+') {
				p++;
				g_string_append(reply, ",J=1");
			} else {
				if (*p == '0' || *p == '-') {
					p++;
				}
				g_string_append(reply, ",J=0");
			}
			break;
		case 'K':
			g_string_append(reply, ",K=1 gypsy");
			break;
		case 'L':
			g_string_append(reply, ",L=3 " VERSION
				" abcdefgijklmnopqrstuvwxyz");
			break;
		case 'M':
			g_string_append_printf(reply, ",M=%d",
				get_fixstatus());
			break;
		case 'N':
			g_string_append(reply, ",N=0");
			break;
		case 'O':
			add_info(reply);
			break;
		case 'P':
			if (isnan(position_latitude)
			 || isnan(position_longitude)) {
				g_string_append(reply, ",P=?");
			} else {
				g_string_append_printf(reply, ",P=%.6f %.6f",
					position_latitude, position_longitude);
			}
			break;
		case 'Q':
			if (isnan(accuracy_pdop)
			 && isnan(accuracy_hdop)
			 && isnan(accuracy_vdop)) {
				g_string_append(reply, ",Q=?");
			} else {
				unsigned i, used;

				for (i = 0, used = 0; i < satellite_satellitec; i++) {
					if (satellite_satellitev[i].used) {
						used++;
					}
				}

				g_string_append_printf(reply,
					",Q=%d %.2f %.2f %.2f %.2f %.2f", used,
					isnan(accuracy_pdop) ? 0.0 : accuracy_pdop,
					isnan(accuracy_hdop) ? 0.0 : accuracy_hdop,
					isnan(accuracy_vdop) ? 0.0 : accuracy_vdop,
					0.0, 0.0);
			}
			break;
		case 'R':
			if (*p == '=') {
				p++;
			}
			if (*p == '2' || *p == '1' || *p == '+') {
				client->raw = TRUE;
				p++;
			} else if (*p == '0' || *p == '-') {
				client->raw = FALSE;
				p++;
			} else {
				client->raw = !client->raw;
			}
			g_string_append_printf(reply, ",R=%1d", client->raw);
			break;
		case 'S':
			g_string_append_printf(reply, ",S=%d",
				MAX(get_fixstatus() - 1, 0));
			break;
		case 'T':
			if (isnan(course_direction)) {
				g_string_append(reply, ",T=?");
			} else {
				g_string_append_printf(reply, ",T=%.4f",
					course_direction);
			}
			break;
		case 'U':
			if (isnan(course_climb)) {
				g_string_append(reply, ",U=?");
			} else {
				g_string_append_printf(reply, ",U=%.3f",
					course_climb);
			}
			break;
		case 'V':
			if (isnan(course_speed)) {
				g_string_append(reply, ",V=?");
			} else {
				g_string_append_printf(reply, ",V=%.3f",
					course_speed * KNOTS_TO_MPS);
			}
			break;
		case 'W':
			if (*p == '=') {
				p++;
			}
			if (*p == '1' || *p == '+') {
				client->watcher = TRUE;
				p++;
			} else if (*p == '0' || *p == '-') {
				client->watcher = FALSE;
				p++;
			} else {
				client->watcher = !client->watcher;
			}
			g_string_append_printf(reply, ",W=%1d", client->watcher);
			break;
		case 'X':
			if (time_seconds <= 0) {
				g_string_append(reply, ",X=?");
			} else {
				g_string_append_printf(reply, ",X=%d.000000",
					time_seconds);
			}
			break;
		case 'Y':
			add_sats(reply);
			break;
		case 'Z':
			g_string_append(reply, ",Z=?"); /* FIXME */
			break;
		case '$':
			g_string_append(reply, ",$=?"); /* FIXME */
			break;
		case '\r':
		case '\n':
			goto finish;
		default:
			break;
		}
	}

finish:
	g_string_append(reply, "\r\n");
	send_to_client(client, reply->str, reply->len);
	g_string_free(reply, TRUE);
	return TRUE;
}

static gboolean
accept_new_client(GIOChannel *source, GIOCondition condition, gpointer data)
{
	struct sockaddr_in sin;
	socklen_t len = sizeof(sin);
	struct client *client;
	int fd;

	g_message("New client request");

	fd = accept(g_io_channel_unix_get_fd(source),
		(struct sockaddr *) &sin, &len);

	if (fd < 0) {
		g_error("Cannot accept new connection: %s", strerror(errno));
	}

	client = g_new0(struct client, 1);

	if (!client) {
		g_error("Cannot allocate memory");
	}

	if (!client_list) {
		g_timeout_add_seconds(1, timeout_handler, NULL);
	}

	if (!client_list && !nowait) {
		start();
	}

	client_list = g_list_prepend(client_list, client);

	client->channel = g_io_channel_unix_new(fd);

	g_io_channel_set_flags(client->channel, G_IO_FLAG_NONBLOCK, NULL);
	g_io_channel_set_encoding(client->channel, NULL, NULL);
	g_io_channel_set_buffered(client->channel, FALSE);

	g_io_add_watch(client->channel,
		G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
		read_from_client, client);

	return TRUE;
}

static void
init_service(const char *service)
{
	struct servent *pse;
	struct sockaddr_in sin;
	GIOChannel *channel;
	int fd, one = 1;

	memset(&sin, 0, sizeof(sin));

	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;

	if (!service) {
		service = getservbyname("gpsd", "tcp") ? "gpsd" : PORT;
	}

	pse = getservbyname(service, "tcp");

	if (pse) {
		sin.sin_port = pse->s_port;
	} else {
		sin.sin_port = htons(atoi(service));
	}

	if (!sin.sin_port) {
		g_error("Cannot get service entry: %s", service);
	}

	fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);

	if (fd < 0) {
		g_error("Cannot create socket: %s", strerror(errno));
	}

	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *) & one, sizeof(one));

	if (bind(fd, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
		g_error("Cannot bind to port %s: %s", service, strerror(errno));
	}

	if (listen(fd, 4) < 0) {
		g_error("Cannot listen on port %s: %s",
			service, strerror(errno));
	}

	channel = g_io_channel_unix_new(fd);

	g_io_channel_set_encoding(channel, NULL, NULL);
	g_io_channel_set_buffered(channel, FALSE);

	g_io_add_watch(channel, G_IO_IN | G_IO_PRI | G_IO_HUP | G_IO_ERR,
		accept_new_client, NULL);
}

/******************************************************************************/

static void
print_to_syslog(const char *string)
{
	syslog(LOG_USER | LOG_INFO, "%s", string);
}

static void
log_to_syslog(
	const char *domain,
	GLogLevelFlags level,
	const char *message,
	gpointer user_data
)
{
	const char *prefix = NULL;
	GString *string;

	if ((level & G_LOG_LEVEL_MASK) == G_LOG_LEVEL_DEBUG) {
		return; /* FIXME */
	}

	switch (level & G_LOG_LEVEL_MASK) {
	case G_LOG_LEVEL_ERROR:    prefix = "ERROR: "; break;
	case G_LOG_LEVEL_CRITICAL: prefix = "CRITICAL: "; break;
	case G_LOG_LEVEL_WARNING:  prefix = "WARNING: "; break;
	case G_LOG_LEVEL_MESSAGE:  prefix = "MESSAGE: "; break;
	case G_LOG_LEVEL_INFO:     prefix = "INFO: "; break;
	case G_LOG_LEVEL_DEBUG:    prefix = "DEBUG: "; break;
	default: break;
	}

	string = g_string_new(prefix);

	if (domain) {
		g_string_append_printf(string, "%s: ", domain);
	}

	g_string_append(string, message);

	print_to_syslog(string->str);

	g_string_free(string, TRUE);
}

/******************************************************************************/

static void
usage(void)
{
	printf("usage: "NAME" [-n] [-N] [-P pidfile] [-S port] [-h]\n\
  Options include:\n\
  -n                        = don't wait for client connects to poll GPS\n\
  -N                        = don't go into background\n\
  -P pidfile                = set file to record process ID\n\
  -S integer (default "PORT") = set port for daemon\n\
  -h                        = help message\n\
  -V                        = emit version and exit.\n");
}

/******************************************************************************/

int
main(int argc, char **argv)
{
	GMainLoop *loop = NULL;
	gboolean background = TRUE;
	char *service = NULL;
	char *pid_file = NULL;

	g_set_prgname(NAME);

	opterr = 0;
	while (1) {
		int option;

		option = getopt(argc, argv, "F:D:S:bhNnP:VR:");

		if (option == -1) {
			break;
		}

		switch (option) {
		case 'D':
			g_warning("Not implemented: -D");
			break;
		case 'F':
			g_warning("Not implemented: -F");
			break;
		case 'N':
			background = FALSE;
			break;
		case 'b':
			g_warning("Not implemented: -b");
			break;
		case 'R':
			g_warning("Not implemented: -R");
			break;
		case 'S':
			service = optarg;
			break;
		case 'n':
			nowait = TRUE;
			break;
		case 'P':
			pid_file = optarg;
			break;
		case 'V':
			printf(NAME" "VERSION"\n");
			exit(EXIT_SUCCESS);
			break;
		case 'h':
		case '?':
		default:
			usage();
			exit(EXIT_SUCCESS);
			break;
		}
	}

	if (background) {
		pid_t pid;
		int fd;

		pid = fork();

		if (pid < 0) {
			g_error("fork() failed");
		} else if (pid > 0) {
			exit(EXIT_SUCCESS);
		}

		openlog(NAME, LOG_PID, LOG_USER);
		g_set_printerr_handler(print_to_syslog);
		g_log_set_default_handler(log_to_syslog, NULL);

		if (setsid() < 0) {
			g_error("setsid() failed");
		}

		chdir("/");

		fd = open("/dev/null", O_RDWR, 0);

		if (fd < 0) {
			g_error("Cannot open /dev/null");
		}

		dup2(fd, STDIN_FILENO);
		dup2(fd, STDOUT_FILENO);
		dup2(fd, STDERR_FILENO);

		close(fd);
	}

	if (pid_file) {
		FILE *fp;

		fp = fopen(pid_file, "w");

		if (!fp) {
			g_error("Cannot create PID file: %s", pid_file);
		}

		fprintf(fp, "%u\n", (unsigned) getpid());
		fclose(fp);
	}

	g_type_init();

	loop = g_main_loop_new(NULL, FALSE);

	nmea_gga = g_string_new(NULL);
	nmea_gsa = g_string_new(NULL);
	nmea_gsv[0] = g_string_new(NULL);
	nmea_gsv[1] = g_string_new(NULL);
	nmea_gsv[2] = g_string_new(NULL);
	nmea_rmc = g_string_new(NULL);
	nmea_gll = g_string_new(NULL);
	nmea_vtg = g_string_new(NULL);

	init_service(service);

	init_dbus();

	if (nowait) {
		start();
	}

	g_main_loop_run(loop);

	return 0;
}

